"""pdt-client command line script."""
import argparse

from . import commands


def main():
    """Entry point to migration data commands."""
    parser = argparse.ArgumentParser(prog="pdt-client")
    parser.add_argument(
        "--username",
        dest="username",
        required=True,
        metavar="USERNAME",
        help="Deployment tool username",
    )
    parser.add_argument(
        "--password",
        dest="password",
        required=True,
        metavar="PASSWORD",
        help="Deployment tool password",
    )
    deployment_url = 'http://deployment.paylogic.eu'
    parser.add_argument(
        "--url",
        dest="url",
        metavar="URL",
        help="Deployment tool url. Defaults to {0}".format(deployment_url),
        required=False,
        default=deployment_url
    )
    subparsers = parser.add_subparsers(help="sub-command help", dest='command')
    subparsers.required = True
    add_subparser_migrate(subparsers)
    add_subparser_migration_data(subparsers)
    add_subparser_case_data(subparsers)
    add_subparser_deploy(subparsers)
    add_subparser_graph(subparsers)
    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)


def add_subparser_migrate(subparsers):
    """Add migrate subparser to the main subparsers collection."""
    parser_migrate = subparsers.add_parser("migrate", help="apply all previously not applied migrations")
    parser_migrate.add_argument(
        "--instance",
        dest="instance",
        metavar="INSTANCE_NAME",
        help="instance for migration",
        required=True,
    )
    phase_options = sorted(commands.MIGRATION_PHASE_MAPPING.keys())
    parser_migrate.add_argument(
        "--phase",
        dest="phase",
        metavar="[{0}]".format(', '.join(phase_options)),
        help="migration phase",
        choices=phase_options,
        required=True,
    )
    parser_migrate.add_argument(
        "--release",
        dest="release",
        metavar="RELEASE_NUMBER",
        help="release name",
        required=True,
    )
    parser_migrate.add_argument(
        "--show",
        dest="show",
        action="store_true",
        help="Only show what to be done, nothing will actually be done",
    )
    parser_migrate.add_argument(
        "--case",
        dest="case",
        type=int,
        metavar="CASE_ID",
        help="case id",
        required=False,
    )
    parser_migrate.add_argument(
        "--connection-string",
        dest="connection_string",
        metavar="CONNECTION_STRING",
        help="connection string",
        required=True,
    )
    parser_migrate.add_argument(
        "--migrations-dir",
        dest="migrations_dir",
        metavar="DIR",
        help="migrations directory",
        required=True,
    )
    parser_migrate.set_defaults(func=lambda args: commands.migrate(
        url=args.url,
        username=args.username,
        password=args.password,
        instance=args.instance,
        phase=args.phase,
        connection_string=args.connection_string,
        migrations_dir=args.migrations_dir,
        release=args.release,
        case=args.case,
        show=args.show)
    )


def add_subparser_migration_data(subparsers):
    """Add migration-data subparser to main subparsers collection."""
    parser_migration_data = subparsers.add_parser("migration-data", help="migration data commands")
    parser_migration_data.add_argument(
        "--case",
        dest="case",
        type=int,
        metavar="CASE_ID",
        help="case id",
        required=False,
    )
    migration_data_subparsers = parser_migration_data.add_subparsers(
        help="sub-command help", dest='migration_data_command')
    parser_push = migration_data_subparsers.add_parser("push", help="push migration data")
    parser_push.add_argument(
        "--alembic-config",
        dest="alembic_config",
        metavar="PATH",
        help="alembic config path",
        required=True,
    )
    parser_push.add_argument(
        "--show",
        dest="show",
        action="store_true",
        help="Only show what to be pushed, nothing will actually be pushed",
    )
    parser_push.set_defaults(func=lambda args: commands.push_data(
        url=args.url,
        username=args.username,
        password=args.password,
        alembic_config=args.alembic_config,
        case=args.case,
        show=args.show)
    )
    parser_get_not_reviewed = migration_data_subparsers.add_parser(
        "get-not-reviewed",
        help="get not reviewed migrations which are in the codebase")
    parser_get_not_reviewed.add_argument(
        "--alembic-config",
        dest="alembic_config",
        metavar="PATH",
        help="alembic config path",
        required=True,
    )
    parser_get_not_reviewed.add_argument(
        "--ci-project",
        dest="ci_project",
        metavar="CI_PROJECT_NAME",
        help="CI project",
        required=True,
    )
    parser_get_not_reviewed.set_defaults(func=lambda args: commands.get_not_reviewed(
        url=args.url,
        username=args.username,
        password=args.password,
        alembic_config=args.alembic_config,
        ci_project=args.ci_project,
        case=args.case)
    )
    parser_get_not_applied = migration_data_subparsers.add_parser(
        "get-not-applied",
        help="get not applied migrations which are in the codebase")
    parser_get_not_applied.add_argument(
        "--instance",
        dest="instance",
        metavar="INSTANCE_NAME",
        help="instance for migration application",
        required=True,
    )
    parser_get_not_applied.add_argument(
        "--release",
        dest="release",
        metavar="RELEASE_NUMBER",
        help="release name",
        required=True,
    )
    parser_get_not_applied.set_defaults(func=lambda args: commands.get_not_applied(
        url=args.url,
        username=args.username,
        password=args.password,
        instance=args.instance,
        release=args.release,
        case=args.case)
    )


def add_subparser_case_data(subparsers):
    """Add case-data subparser to main subparsers collection."""
    parser_case_data = subparsers.add_parser("case-data", help="case data commands")
    parser_case_data.add_argument(
        "--case",
        dest="case",
        type=int,
        metavar="CASE_ID",
        help="case id",
        required=False,
    )
    migration_data_subparsers = parser_case_data.add_subparsers(
        help="sub-command help", dest='case_data_command')
    parser_get_revisions = migration_data_subparsers.add_parser(
        "get-not-deployed",
        help="get not deployed case ids, revisions and titles")
    parser_get_revisions.add_argument(
        "--ci-project",
        dest="ci_project",
        metavar="CI_PROJECT_NAME",
        help="CI project",
        required=True,
    )
    parser_get_revisions.add_argument(
        "--release",
        dest="release",
        metavar="RELEASE_NUMBER",
        help="release name",
        required=True,
    )
    parser_get_revisions.add_argument(
        "--instance",
        dest="instance",
        metavar="INSTANCE_NAME",
        help="instance for deployment",
        required=True,
    )
    parser_get_revisions.set_defaults(func=lambda args: commands.get_not_deployed_cases(
        url=args.url,
        username=args.username,
        password=args.password,
        ci_project=args.ci_project,
        release=args.release,
        instance=args.instance,
        case=args.case)
    )


def add_subparser_deploy(subparsers):
    """Add deploy subparser to main subparsers collection."""
    parser_deploy = subparsers.add_parser("deploy", help="report the deployment results")
    parser_deploy.add_argument(
        "--instance",
        dest="instance",
        metavar="INSTANCE_NAME",
        help="instance for deployment",
        required=True,
    )
    parser_deploy.add_argument(
        "--status",
        dest="status",
        metavar="[dpl,err]",
        choices=('dpl', 'err'),
        help="deployment status, pass 'dpl' for 'Deployed OK', 'err' for 'Deployment Error'",
        required=True,
    )
    parser_deploy.add_argument(
        "--case",
        dest="cases",
        action="append",
        metavar="CASE_ID",
        type=int,
        default=[],
        help="id of the deployed case",
        required=False,
    )
    parser_deploy.add_argument(
        'log',
        type=argparse.FileType('r'),
        default='-',
    )
    parser_deploy.set_defaults(func=lambda args: commands.deploy(
        url=args.url,
        username=args.username,
        password=args.password,
        instance=args.instance,
        status=args.status,
        cases=args.cases,
        log=args.log)
    )


def add_subparser_graph(subparsers):
    """Add a graph subparser to main subparser collection."""
    parser_graph = subparsers.add_parser("graph", help="Graph the migrations")
    parser_graph.add_argument(
        '--filename',
        dest="filename",
        metavar="FILENAME",
        help="The filename to store the dotfile too.",
        required=True,
    )
    parser_graph.add_argument(
        "--alembic-config",
        dest="alembic_config",
        metavar="PATH",
        help="alembic config path",
        required=True,
    )
    parser_graph.add_argument(
        '--quiet',
        dest='verbose',
        action='store_false',
        help="Should it print output?",
        required=False
    )

    parser_graph.set_defaults(func=lambda args: commands.graph(
        url=args.url,
        username=args.username,
        password=args.password,
        alembic_config=args.alembic_config,
        filename=args.filename,
        verbose=args.verbose)
    )



"""pdt-client commands."""
import json
from functools import partial
import os
import pprint
import sys
import traceback

from alembic.config import Config
from alembic_offline import get_migrations_data, generate_migration_graph
from capturer import CaptureOutput
import requests
import six
import sqlalchemy

try:
    import subprocess32 as subprocess
except ImportError:  # pragma: no cover
    import subprocess


MIGRATION_PHASE_MAPPING = {
    'before-deploy': 'pre_deploy_steps',
    'after-deploy': 'post_deploy_steps',
    'final': 'final_steps',
}


def apply_migration_step(
        url, username, password, migration, step, instance, phase, engine, migrations_dir, show):
    """Apply migration step."""
    print("-- Applying migration step: id={step[id]}, position={step[position]}".format(step=step))
    report = ''
    exc = None
    try:
        if step['type'] == engine.dialect.name:
            if show:
                print(step['code'])
            else:
                report = 'Executed SQL with rowcount: {0}'.format(engine.execute(step['code']).rowcount)
        else:
            path = os.path.join(migrations_dir, step['path'])
            args = [path]
            if step['type'] == 'python':
                args.insert(0, sys.executable)
            if show:
                print('-- Script to be executed: {0}'.format(' '.join(args)))
            else:
                with CaptureOutput() as capturer:
                    subprocess.check_call(args)
                    report = capturer.get_text()
        status = 'apl'
    except Exception as exc:
        report = traceback.format_exc()
        status = 'err'
    try:
        if not show:
            print('Applied migration step with status: {0}'.format(status))
            print('Reporting status to deployment tool')
            data = {
                "report": {
                    "migration": {
                        "uid": migration['uid']
                    },
                    "instance": {
                        "name": instance,
                    }
                },
                "step": {
                    "id": step['id']
                },
                "status": status,
                "log": report
            }
            response = requests.post(
                '{url}/api/migration-step-reports/'.format(url=url),
                auth=(username, password),
                data=json.dumps(data, sort_keys=True),
                headers={'content-type': 'application/json'}
            )
            try:
                response.raise_for_status()
            except Exception:
                pprint.pprint(response.json())
                raise
    finally:
        if exc:
            six.reraise(exc, None, sys.exc_traceback)


def apply_migration(url, username, password, migration, instance, phase, engine, migrations_dir, show):
    """Apply migration."""
    print("-- Applying migration: {migration[uid]}".format(migration=migration))
    for step in migration[MIGRATION_PHASE_MAPPING[phase]]:
        apply_migration_step(
            url=url, username=username, password=password,
            migration=migration, step=step, instance=instance, phase=phase, engine=engine,
            migrations_dir=migrations_dir, show=show)


def migrate(
        url, username, password, instance, phase, connection_string, migrations_dir, release, case=None,
        show=False):
    """Apply previously not applied migrations."""
    engine = sqlalchemy.create_engine(connection_string)
    params = dict(
        reviewed=True, exclude_status='apl', instance=instance, release=release)
    if case:
        params['case'] = case
    response = requests.get(
        '{url}/api/migrations/'.format(url=url),
        auth=(username, password),
        params=params,
        headers={'content-type': 'application/json'}
    )
    try:
        response.raise_for_status()
    except Exception:
        pprint.pprint(response.json())
        raise
    migrations = response.json()
    print('-- Got migration data with count: {0}'.format(len(migrations)))
    for migration in migrations:
        apply_migration(
            url=url, username=username, password=password,
            migration=migration, instance=instance, phase=phase, engine=engine,
            migrations_dir=migrations_dir, show=show)


def get_phase_steps(migration, phase):
    """Get migration phase steps.

    :param migration: migration data dict
    :type migration: dict
    :param phase: migration phase name
    :type phase: str

    :return: generator of migration step data in form: {'position': 1, code: 'code', 'type': 'type'[, 'path': 'path']}
    """
    for index, step in enumerate(migration['phases'].get(phase, dict(steps=[]))['steps']):
        kwargs = {}
        if 'path' in step:
            kwargs['path'] = step['path']
        yield dict(position=index, code=step['script'], type=step['type'], **kwargs)


def push_data(url, username, password, alembic_config, case=None, show=False):
    """Push migration data.

    :args: command line arguments namespace object

    :raises: Exception if PDT replied with an error
    """
    config = Config(alembic_config)
    for migration in get_migrations_data(config):
        if case == migration['attributes']['case_id'] or not case:
            print(
                'Got migration data for migration: {migration[revision]}, case: {migration[attributes][case_id]}'
                .format(migration=migration))
            call_url = '{0}/api/migrations/'.format(url)
            data = {
                "uid": migration['revision'],
                "parent": migration['down_revision'],
                "case": {
                    "id": str(migration['attributes']['case_id'])
                },
                "pre_deploy_steps": list(get_phase_steps(migration, 'before-deploy')),
                "post_deploy_steps": list(get_phase_steps(migration, 'after-deploy')),
                "final_steps": list(get_phase_steps(migration, 'final')),
            }
            if show:
                print('URL: {call_url}, data: \n{data}'.format(call_url=call_url, data=pprint.pformat(data)))
            else:
                response = requests.post(
                    call_url,
                    auth=(username, password),
                    data=json.dumps(data, sort_keys=True), headers={'content-type': 'application/json'})
                try:
                    response.raise_for_status()
                    print(
                        'Pushed migration data for migration: {migration[revision]}, '
                        'case: {migration[attributes][case_id]}'
                        .format(
                            migration=migration))
                except Exception:
                    pprint.pprint(response.json())
                    raise
            if case:
                break


def get_not_reviewed(url, username, password, alembic_config, ci_project, case=None):
    """Get not reviewed migrations.

    :args: command line arguments namespace object

    :raises:
        * Exception - PDT replied with an error
        * SystemExit(<number>) - found <number> of not reviewed migrations
    """
    config = Config(alembic_config)
    migrations = frozenset(migration['attributes']['case_id'] for migration in get_migrations_data(config))
    params = dict(reviewed=True, ci_project=ci_project)
    if case:
        params['case'] = case
    response = requests.get(
        '{0}/api/migrations/'.format(url),
        params=params,
        auth=(username, password),
        headers={'content-type': 'application/json'}
    )
    try:
        response.raise_for_status()
        print('Got migration data')
    except Exception:
        pprint.pprint(response.json())
        raise
    response_migrations = frozenset(migration['case']['id'] for migration in response.json())
    diff = migrations - response_migrations
    if diff:
        print('Found not reviewed migrations for these cases: {0}'.format(sorted(diff)))
        sys.exit(len(diff))
    else:
        print('No not reviewed migrations found so far')


def get_not_applied(url, username, password, instance, release, case=None):
    """Get not applied migrations.

    :args: command line arguments namespace object

    :raises:
        * Exception - PDT replied with an error
        * SystemExit(<number>) - found <number> of not applied migrations
    """
    params = dict(
        exclude_status='apl', instance=instance, release=release)
    if case:
        params['case'] = case
    response = requests.get(
        '{0}/api/migrations/'.format(url),
        params=params,
        auth=(username, password),
        headers={'content-type': 'application/json'}
    )
    try:
        response.raise_for_status()
        print('Got migration data')
    except Exception:
        pprint.pprint(response.json())
        raise
    response_migrations = sorted(migration['case']['id'] for migration in response.json())
    if response_migrations:
        print('Found not applied migrations for these cases: {0}'.format(response_migrations))
        sys.exit(len(response_migrations))
    else:
        print('No not applied migrations found so far')


def get_not_deployed_cases(url, username, password, ci_project, release, instance, case=None):
    """Get not deployed cases.

    :args: command line arguments namespace object

    :raises:
        * Exception - PDT replied with an error
        * SystemExit(<number>) - found <number> of not applied migrations
    """
    params = dict(ci_project=ci_project, release=release, exclude_deployed_on=instance)
    if case:
        params['id'] = case
    response = requests.get(
        '{0}/api/cases/'.format(url),
        params=params,
        auth=(username, password),
        headers={'content-type': 'application/json'}
    )
    try:
        response.raise_for_status()
        for case in response.json():
            print('{case[id]}\t{case[revision]}\t{case[title]}'.format(case=case))
    except Exception:
        pprint.pprint(response.json())
        raise


def deploy(url, username, password, instance, status, log, cases):
    """Report the deployment."""
    data = dict(
        status=status,
        instance=dict(name=instance),
        cases=[dict(id=case) for case in cases],
        log=log.read(),
    )
    response = requests.post(
        '{url}/api/deployment-reports/'.format(url=url),
        auth=(username, password),
        data=json.dumps(data, sort_keys=True),
        headers={'content-type': 'application/json'}
    )
    try:
        response.raise_for_status()
        print(
            'Reported the deployment for instance: {instance[name]}, cases: {cases}'
            .format(**data))
    except Exception:
        pprint.pprint(response.json())
        raise


def _label_callback(data, release_numbers=None):
    """Generate a label for the revision.

    :param data: Dict with migration data.
    :type data: dict.

    :param release_numbers: dict containing the migration data from pdt.
    :type release_numbers: dict.

    :return: String with the label.
    :rtype: str
    """
    if not release_numbers:
        release_numbers = {}
    attributes = []
    release = release_numbers.get(data['revision'], "Unknown")
    attributes.append(u'- Release: {0}'.format(release))
    for key, value in data['attributes'].items():
        attributes.append(u'- {0}: {1}'.format(key, value))
    return u'{0}\n{1}'.format(data['revision'], '\n'.join(attributes))


def graph(url, username, password, alembic_config, filename, verbose=True):
    """Generate a dotfile with an overview of all the migrations."""
    config = Config(alembic_config)

    response = requests.get(
        '{0}/api/migrations/'.format(url),
        auth=(username, password),
        headers={'content-type': 'application/json'}
    )
    release_numbers = {}
    try:
        release_numbers = dict(
            (migration['uid'], (migration.get('release') or {}).get('number'))
            for migration in response.json()
        )
    except KeyError:
        pass

    label_callback = partial(_label_callback, release_numbers=release_numbers)

    with open(filename, 'w') as fp:
        fp.write(generate_migration_graph(config, label_callback))

    if verbose:
        print("Done")
        print("To generate an image use: dot -Tpng -O {0}".format(filename))



"""Setuptools entry point."""

import codecs
from os.path import abspath, dirname, join
import sys

from setuptools import setup, find_packages

from pdt_client import __version__

long_description = []

install_requires = [
    'alembic',
    'alembic-offline>=1.2.0',
    'capturer>=2.1',
    'requests',
    'six',
    'sqlalchemy',
]

if sys.version_info < (3, 0, 0):
    install_requires.append('subprocess32')

tests_require = [
    "mock==1.0.1",
    "pylama==6.3.1",
    "pytest==2.7.0",
    "pytest-mock==0.4.3"
]

for text_file in ['README.rst', 'CHANGES.rst']:
    with codecs.open(join(dirname(abspath(__file__)), text_file), encoding='utf-8') as f:
        long_description.append(f.read())

setup(
    name="pdt-client",
    description="Client for paylogic deployment tool",
    long_description='\n'.join(long_description),
    author="Anatoly Bubenkov, Paylogic International and others",
    license="MIT license",
    author_email="developers@paylogic.com",
    url="https://github.com/paylogic/pdt-client",
    version=__version__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    packages=find_packages(include=["pdt_client*"]),
    install_requires=install_requires,
    entry_points="""
    [console_scripts]
    pdt-client = pdt_client.script:main
    """,
    tests_require=tests_require,
    extras_require=dict(test=tests_require)
)
